from collections import Counter
from flask import session
import cqi
import json
import math
import os
from app import db, socketio
from app.decorators import socketio_login_required
from app.models import Corpus
from . import NAMESPACE as ns
from .utils import cqi_over_socketio, lookups_by_cpos


@socketio.on('cqi.corpora.corpus.drop', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_drop(cqi_client: cqi.CQiClient, corpus_name: str):
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_status = cqi_corpus.drop()
    payload = {'code': cqi_status.code,
               'msg': cqi_status.__class__.__name__}
    return {'code': 200, 'msg': 'OK', 'payload': payload}


@socketio.on('cqi.corpora.corpus.query', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_query(cqi_client: cqi.CQiClient, corpus_name: str, subcorpus_name: str, query: str):  # noqa
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_status = cqi_corpus.query(subcorpus_name, query)
    payload = {'code': cqi_status.code,
               'msg': cqi_status.__class__.__name__}
    return {'code': 200, 'msg': 'OK', 'payload': payload}


###############################################################################
# nopaque specific CQi extensions                                             #
###############################################################################
@socketio.on('cqi.corpora.corpus.update_db', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_update_db(cqi_client: cqi.CQiClient, corpus_name: str):
    corpus = Corpus.query.get(session['d']['corpus_id'])
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    corpus.num_tokens = cqi_corpus.size
    db.session.commit()


@socketio.on('cqi.corpora.corpus.get_visualization_data', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_get_visualization_data(cqi_client: cqi.CQiClient, corpus_name: str):
    corpus = Corpus.query.get(session['d']['corpus_id'])
    visualization_data_file_path = os.path.join(corpus.path, 'cwb', 'visualization_data.json')
    if os.path.exists(visualization_data_file_path):
        with open(visualization_data_file_path, 'r') as f:
            payload = json.load(f)
        return {'code': 200, 'msg': 'OK', 'payload': payload}
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    ##########################################################################
    # A faster way to get cpos boundaries for smaller s_attrs                #
    ##########################################################################
    # cqi_corpus.query('Last', '<s> []* </s>;')
    # cqi_subcorpus = cqi_corpus.subcorpora.get('Last')
    # print(cqi_subcorpus.size)
    # first_match = 0
    # last_match = cqi_subcorpus.attrs['size'] - 1
    # match_boundaries = zip(
    #     list(range(first_match, last_match + 1)),
    #     cqi_subcorpus.dump(cqi_subcorpus.attrs['fields']['match'], first_match, last_match),
    #     cqi_subcorpus.dump(cqi_subcorpus.attrs['fields']['matchend'], first_match, last_match)
    # )
    # for x in match_boundaries:
    #     print(x)
    cqi_p_attrs = {
        p_attr.name: p_attr
        for p_attr in cqi_corpus.positional_attributes.list()
    }
    cqi_s_attrs = {
        s_attr.name: s_attr
        for s_attr in cqi_corpus.structural_attributes.list()
    }
    payload = {
        'corpus': {
            'bounds': [0, cqi_corpus.size - 1],
            'counts': {
                'token': cqi_corpus.size
            },
            'freqs': {}
        },
        'p_attrs': {},
        's_attrs': {},
        'values': {'p_attrs': {}, 's_attrs': {}}
    }
    for p_attr in cqi_p_attrs.values():
        payload['corpus']['freqs'][p_attr.name] = dict(
            zip(
                range(0, p_attr.lexicon_size),
                p_attr.freqs_by_ids(list(range(0, p_attr.lexicon_size)))
            )
        )
        payload['p_attrs'][p_attr.name] = dict(
            zip(
                range(0, cqi_corpus.size),
                p_attr.ids_by_cpos(list(range(0, cqi_corpus.size)))
            )
        )
        payload['values']['p_attrs'][p_attr.name] = dict(
            zip(
                range(0, p_attr.lexicon_size),
                p_attr.values_by_ids(list(range(0, p_attr.lexicon_size)))
            )
        )
    for s_attr in cqi_s_attrs.values():
        if s_attr.has_values:
            continue
        payload['corpus']['counts'][s_attr.name] = s_attr.size
        payload['s_attrs'][s_attr.name] = {'lexicon': {}, 'values': None}
        payload['values']['s_attrs'][s_attr.name] = {}
        for id in range(0, s_attr.size):
            payload['s_attrs'][s_attr.name]['lexicon'][id] = {}
            lbound, rbound = s_attr.cpos_by_id(id)
            payload['s_attrs'][s_attr.name]['lexicon'][id]['bounds'] = [lbound, rbound]
            payload['s_attrs'][s_attr.name]['lexicon'][id]['counts'] = {}
            payload['s_attrs'][s_attr.name]['lexicon'][id]['counts']['token'] = rbound - lbound + 1
            if s_attr.name not in ['text', 's']:
                continue
            cpos_range = range(lbound, rbound + 1)
            payload['s_attrs'][s_attr.name]['lexicon'][id]['counts']['ent'] = len({x for x in cqi_s_attrs['ent'].ids_by_cpos(list(cpos_range)) if x != -1})
            if s_attr.name != 'text':
                continue
            payload['s_attrs'][s_attr.name]['lexicon'][id]['counts']['s'] = len({x for x in cqi_s_attrs['s'].ids_by_cpos(list(cpos_range)) if x != -1})
            payload['s_attrs'][s_attr.name]['lexicon'][id]['freqs'] = {}
            for p_attr in cqi_p_attrs.values():
                payload['s_attrs'][s_attr.name]['lexicon'][id]['freqs'][p_attr.name] = dict(Counter(p_attr.ids_by_cpos(list(cpos_range))))
        sub_s_attrs = cqi_corpus.structural_attributes.list(filters={'part_of': s_attr})
        s_attr_value_names = [
            sub_s_attr.name[(len(s_attr.name) + 1):]
            for sub_s_attr in sub_s_attrs
        ]
        sub_s_attr_values = [
            sub_s_attr.values_by_ids(list(range(0, s_attr.size)))
            for sub_s_attr in sub_s_attrs
        ]
        payload['s_attrs'][s_attr.name]['values'] = s_attr_value_names
        payload['values']['s_attrs'][s_attr.name] = {
            s_attr_id: {
                s_attr_value_name: sub_s_attr_values[s_attr_value_name_idx][s_attr_id_idx]
                for s_attr_value_name_idx, s_attr_value_name in enumerate(
                    payload['s_attrs'][s_attr.name]['values']
                )
            } for s_attr_id_idx, s_attr_id in enumerate(range(0, s_attr.size))
        }
    with open(visualization_data_file_path, 'w') as f:
        json.dump(payload, f)
    return {'code': 200, 'msg': 'OK', 'payload': payload}


@socketio.on('cqi.corpora.corpus.paginate', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_paginate(cqi_client: cqi.CQiClient, corpus_name: str, page: int = 1, per_page: int = 20):  # noqa
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    # Sanity checks
    if (
        per_page < 1
        or page < 1
        or (
            cqi_corpus.size > 0
            and page > math.ceil(cqi_corpus.size / per_page)
        )
    ):
        return {'code': 416, 'msg': 'Range Not Satisfiable'}
    first_cpos = (page - 1) * per_page
    last_cpos = min(cqi_corpus.size, first_cpos + per_page)
    cpos_list = [*range(first_cpos, last_cpos)]
    lookups = lookups_by_cpos(cqi_corpus, cpos_list)
    payload = {}
    # the items for the current page
    payload['items'] = [cpos_list]
    # the lookups for the items
    payload['lookups'] = lookups
    # the total number of items matching the query
    payload['total'] = cqi_corpus.size
    # the number of items to be displayed on a page.
    payload['per_page'] = per_page
    # The total number of pages
    payload['pages'] = math.ceil(payload['total'] / payload['per_page'])
    # the current page number (1 indexed)
    payload['page'] = page if payload['pages'] > 0 else None
    # True if a previous page exists
    payload['has_prev'] = payload['page'] > 1 if payload['page'] else False
    # True if a next page exists.
    payload['has_next'] = payload['page'] < payload['pages'] if payload['page'] else False  # noqa
    # Number of the previous page.
    payload['prev_num'] = payload['page'] - 1 if payload['has_prev'] else None
    # Number of the next page
    payload['next_num'] = payload['page'] + 1 if payload['has_next'] else None
    return {'code': 200, 'msg': 'OK', 'payload': payload}
